# Parrot post processor -- Jim Yehle -- Last modified: Mar 16, 2008

import sys, math

PARROT_POST_VERSION = 1.6

PRE_LAUNCH,BOOST,COAST,DESCENT,DOWN = range(5) # a la Enum
NOM_VREF = 1.8 # Nominal reference voltage
G_THRESH_DEFAULT = 2.5 # which is 1.5 vertical Gs plus 1 for gravity
G_THRESH_COASTING = -0.1 # compare axial accel against this to detect burnout
DECAY=.85 # exponential decay factor: 0=no filtering, 1=infinite
SLOW_DECAY=.3 # less filtering when sampling rate drops (1Hz after 308 sec)
DESC_RATE_DECAY = .9 # this value when 2-sec window is 100 samples 
DESC_RATE_SLOW_DECAY = .2 # this value when 2-sec window is 2 samples
G_FILT_DECAY = .75 # filter factor for Accel plot; launch detect is unfiltered
DOWN_THESH = 200 # above this alt, don't run arrived-on-ground test
DESC_WINDOW_SEC = 2 # #-of-seconds backward to look, in descent delta math
DOWN_DESC_RATE_THRESH = .5 # ft/sec .. DOWN test true if descent rate < this
DOWN_THRESH_SEC = 2 # sec .. DOWN test must hold off this long after launch
VERBOSE_DEFAULT = 1
EARTH_G = 32.174 # ft/sec/sec = 1 G
STD_ATM_LAPSE_RATE = 0.00198*9./5. # deg F/foot .. for baro temp compensation

parrot_prog = [ ('Default as shipped' ,            "vhrJvZ@L@rNT"),
                ('2nd Stage at Burnout' ,          "vhrJ^n@L@rNT"),
                ('2nd Stage with Vertical Check' , "vhrJlb@L@rNT"),
                ('Timer, Output 3, 6 sec' ,   "vhrJvnBJ@rNT"),
                ('Timer, Output 3, 8 sec' ,   "vhrJvnBd@rNT"),
                ('Timer, Output 3, 10 sec' ,  "vhrJvnB|@rNT"), # Template wrong .. this sets all 3 continuous outputs
                ('Timer, Output 3, 10 sec, continuous 3' ,  "vhrJv~B|@rNT"),
                ('Timer, Output 3, 12 sec' ,  "vhrJvnDV@rNT"),
                ('Timer, Output 3, 14 sec' ,  "vhrJvnDn@rNT"),
                ('Output 3 Disabled' ,             "vhrJfF@L@rNT")
              ]

# Interpretations of bits of FER (Flight Event Register)
UPPER_BYTE = 256
#FER_BITMASK_LIFTOFF_DETECTED = 128 * UPPER_BYTE # Signed!! sheesh.
FER_BITMASK_ACCEL_LT_ZERO = 32 * UPPER_BYTE
FER_BITMASK_TIME_LT_TVAL = 16 * UPPER_BYTE
FER_BITMASK_TIME_GT_TVAL = 8 * UPPER_BYTE
FER_BITMASK_DELTA_PRESS_LT_PVAL1 = 4 * UPPER_BYTE
FER_BITMASK_DELTA_PRESS_GT_PVAL2 = 2 * UPPER_BYTE
FER_BITMASK_PRESS_INCREASING = 32
FER_BITMASK_LATCH_OUTPUT = 16
FER_BITMASK_PRESS_DECREASING = 8
FER_BITMASK_VEL_LT_400MPH = 4
FER_BITMASK_VEL_LT_0 = 2

FER_ACCEL_APOGEE = (FER_BITMASK_TIME_GT_TVAL \
                     + FER_BITMASK_VEL_LT_400MPH \
                     + FER_BITMASK_VEL_LT_0)

FER_BURNOUT = (FER_BITMASK_ACCEL_LT_ZERO)
               
FER_LOW_ALT_DEPLOY = (FER_BITMASK_TIME_GT_TVAL \
                       + FER_BITMASK_DELTA_PRESS_LT_PVAL1 \
                       + FER_BITMASK_PRESS_INCREASING \
                       + FER_BITMASK_VEL_LT_400MPH)

FER_BARO_APOGEE = (FER_BITMASK_TIME_GT_TVAL \
                    + FER_BITMASK_PRESS_INCREASING \
                    + FER_BITMASK_VEL_LT_400MPH \
                    + FER_BITMASK_LATCH_OUTPUT)

# ----- csv -------------------------------------------------------------------
#

def csv(arg,OFS):

  """If string contain field-separator chars, they must be quoted
     CSV is really a misnomer, since this should work
     for other field separators besides commas (e.g. tabs)
     (Isn't there a standard CSV module in Python now?!!)"""

  ret = str(arg)
  if OFS in ret: # I think you could also handle \n chars in fields here also
    if '"' in ret:
       # Ugh. Emit "" for each "
       ret = ''
       for ch in str(arg):
         if ch == '"':  ret += '""'
         else:          ret += ch
    return '"' + ret + '"'
  return ret


# ----- print1line ------------------------------------------------------------
#

# Indices of print-buffer fld[]
N_COLS = 20 # number of fields in fld[]
# note: the first 8 match what comes out of the Parrot
SEX_IX,AXIAL_G_IX,LATERAL_G_IX,VBAT_IX,VREF_IX,TEMP_IX,ATM_IX,IND_IX,\
 SEX_ADJ_IX,G_AX_ADJ_IX,G_LAT_ADJ_IX,ATM_ADJ_IX,BARO_ASL_ALT_IX,\
 BARO_AGL_ALT_IX,MAX_IX,FILT_ALT_IX,DOWN_IX,FILT_ACCEL_IX,INTACC_VEL_IX,\
 TCOMP_BARO_ALT = range(N_COLS)
PYRO_CONT_IX = VBAT_IX # Parrot v.2: fld[4] is pyro continuity voltage
FER_IX = VREF_IX       # Parrot v.2: fld[4] is Flight Event Register

def print1line(fld,outfile,behavior):

  """Given a list of fields, print them, separated by field-separator.
     In spreadsheet mode, dumps them all; otherwise, be more selective.
     If no outfile, skips writing it."""

  if outfile:
    OFS = behavior['OFS']
    outfile.write( OFS.join( csv(x,OFS) for x in [
     fld[SEX_ADJ_IX],fld[G_AX_ADJ_IX],fld[G_LAT_ADJ_IX],fld[TEMP_IX],
      fld[ATM_ADJ_IX],fld[BARO_ASL_ALT_IX],fld[BARO_AGL_ALT_IX],
      fld[FILT_ALT_IX],fld[MAX_IX],fld[DOWN_IX],fld[FILT_ACCEL_IX],
      fld[FER_IX],fld[PYRO_CONT_IX],fld[TCOMP_BARO_ALT]])\
       + "\n")

class Accum:
  def __init__(self):
    self.x = []
    self.t = []
  def append(self,this_x,this_t):
    self.x.append(this_x)
    self.t.append(this_t)

class Flight:
  def __init__(self):
    self.liftoff_time = 0.0  #sec
    self.liftoff_alt = 0     # calculate this when launch detected
    self.ambient_pad_temp = 0# calculate this when launch detected
    self.alt = Accum()       # All baro alt readings, unfiltered
    self.filt = Accum()      # All baro alt readings, filtered
    self.time_hist = []      # Corresponding times
    self.desc = Accum()      # rate-of-descent history
    self.desc_rate = 0
    self.accel_hist = []     # filtered axial acceleration, in Gs
    self.liftoff_time_ix = 0 # how many samples in *_hist[] before liftoff
    self.baro_alt_max = 0
    self.pre_launch = []     # list of fld[]s which need liftoff_alt fixup
    self.max_alt_time = 0    # when apogee happened (after takeoff)
    self.phase = PRE_LAUNCH
    self.ofname = None       # output file name
    self.filt_alt = 0.0      # filtered barometric altitude AGL
    self.filt_accel = 0.0
    self.avg_thrust_accel = 0
    self.burnout_time = 0
    self.n_samples = 0
    self.max_accel = 0
    self.temp = Accum()      # temperatures
    self.intacc_alt = 0      # integrated-accelerometer altitude
    self.acc_alt = Accum()   #  .. and a list of that
    self.tcomp_alt = Accum() # temperature-compensated baro alt, AGL ft
    self.tcomp_agl = ''
    self.tcomp_liftoff_alt = 0  # calculate this when launch detected
    self.delta_temp_from_std_atm = 0
    self.intacc_vel = 0      # integrated-accelerometer velocity
    self.acc_vel = Accum()   #  .. a list of that (negative, to match descent)
    self.deploy_vel_acc = 0  # velocity at deployment, integrated accelerometer
    self.deploy_vel_baro = 0 # velocity at deployment, differentiated barometer
    self.down = ''           # descent rate, f/s
    self.landing_alt = 0     # baro altitude at landing
    self.FER = Accum()       # flight-event-register
    self.FER_accel_apo = 0
    self.FER_accel_apo_time = None
    self.FER_baro_apo = 0
    self.FER_baro_apo_time = None
    self.FER_low_alt_deploy = 0
    self.FER_low_alt_deploy_time = None
    self.FER_liftoff_time = None
    self.main_deploy_alt = 0
    self.FER_burnout_time = None
    self.Vref = NOM_VREF # Voltage reference (sampled by Parrot v1;not v2)
    self.mux = None # A/D multiplexer chooses Vcont,FER,Temp,Baro,Vcont,...
                    # when? Gotta figure out ..
                    # n_samples==mux is when Vcont first sampled,
                    # n_samples+1 for FER, then temp, then baro, vcont, ...
    self.desc_window = DESC_WINDOW_SEC*50 # init: time -> #-of-readings
    self.vcont = ''
    self.vbat = ''

  # ----- fer_events ----------------------------------------------------------
  #

  def fer_events( self, FER, t):

    """Identify flight event times,
       based on bits in the Flight Events Register.
       Also latch velocity at deployment (accel and baro).
       Returns string of anything interesting to be logged."""

    ret = None

    if FER == '': return ret # no valid data yet
    if FER >=0:  return ret # liftoff not detected, so nothing else matters

    if self.FER_liftoff_time == None:
      self.FER_liftoff_time = t
      # Transition of self.phase to BOOST happens elsewhere
    if self.FER_burnout_time == None:
      if (FER & FER_BURNOUT) == FER_BURNOUT:
        self.FER_burnout_time = t
        # Transition of self.phase to COAST happens elsewhere
    if self.FER_accel_apo_time == None:
      if (FER & FER_ACCEL_APOGEE) == FER_ACCEL_APOGEE:
        self.FER_accel_apo_time = t # time of deployment initiation
        self.deploy_vel_acc = self.intacc_vel # velocity at deployment (accel)
        if self.phase != DESCENT:
          self.phase = DESCENT
          ret = "Phase -> Descent, T+ %g sec" % t
    if self.FER_baro_apo_time == None:
      if (FER & FER_BARO_APOGEE) == FER_BARO_APOGEE:
        self.FER_baro_apo_time = t
        # Capture velocity at deployment (baro)
        # This isn't very accurate, since it looks back a whole 2 seconds.
        self.deploy_vel_baro = -self.down
        if self.phase != DESCENT:
          self.phase = DESCENT
          ret = "Phase -> Descent, T+ %g sec" % t
    if self.FER_low_alt_deploy_time == None:
      if (FER & FER_LOW_ALT_DEPLOY) == FER_LOW_ALT_DEPLOY:
        self.FER_low_alt_deploy_time = t
        # Latch alt, to calculate final-descent speed
        self.main_deploy_alt = self.filt_alt

    return ret


  # ----- temp_comp_baro ------------------------------------------------------
  #

  def temp_comp_baro( self, baro_asl):
    """Temperature compensation of baro readings.
       Requires pre-launch average temperature, temp-adjusted liftoff-altitude,
       and delta_temp_from_std_atm to be known before results are meaningful.
       Returns alt offset adjustment."""

    lapsed = STD_ATM_LAPSE_RATE*baro_asl # temp var, to save a multiplication
    return (self.delta_temp_from_std_atm/STD_ATM_LAPSE_RATE) \
            * math.log(1.+lapsed/(519.0+lapsed))


  # ----- post_process_line ---------------------------------------------------
  #

  def post_process_line( self, linebuf, outfile, err, log, behavior):

    """Perform post-processing on Parrot raw flight data.
       String linebuf is input.  Func err(text) reports errors;  func
       log() gives info.  Tolerates None for outfile, log(), or err() 
       if no output desired."""

    # Input, Parrot v1 e.g.
    # 1712.010	77	2	3818	1800	781	8162	4
    try:
      fld = linebuf.split() # cut up @ whitespace (leading SPs, tab separators)
    except:
      if err!=None: err("Could not parse line: %s" % repr(linebuf) )
      return
    if len(fld) != 8 and err!=None: # 8 columns
      err("Could not parse line as 8 fields: %s (skipping)" % repr(linebuf) )
      return
    self.n_samples += 1
    sex = float(fld[SEX_IX])      # [ss]s.sss time since log start, NOT liftoff
    axial_g = float(fld[AXIAL_G_IX])/100   # hundredths of a G -> Gs
    lateral_g = float(fld[LATERAL_G_IX])/100 # ditto
    temp = float(fld[TEMP_IX])/10   # raw is tenths of a degrees F
    if behavior['parrot_version']<2: # updated mux-mod==1; no harm now also
      self.Vref = float(fld[VREF_IX])/1000 # Ref voltage .. mV -> Volts
    mod = temp_adj_asl_ft = None
    FER = max = ''
    # Parrot doesn't sample all sensors at full speed
    if self.mux:
      mod = (self.n_samples-self.mux) % 4 # which multiplexer selection now?

    sex_adj = sex - self.liftoff_time
    g_ax_adj = axial_g*NOM_VREF/self.Vref # Axial G-sensor, adjusted
    g_lat_adj = lateral_g*NOM_VREF/self.Vref # Lateral G-sensor, adjusted

    # Apply digital filters: unit-delay*DECAY + current-value*(1-DECAY)

    # Integrated-accelerometer altitude, every sample
    try:
      delta_t = sex_adj - self.time_hist[-1] # time tween prior two samples
    except IndexError: # first time: self.time_hist has no data yet
      delta_t = 0
    if delta_t > 0:
      # Get average acceleration between this reading and prior, in ft/sec/sec
      avg_accel = EARTH_G * (self.filt_accel + self.accel_hist[-1]) / 2.0
    else:
      avg_accel = EARTH_G * self.filt_accel / 2.0 # assume prior G reading = 0
      delta_t = .05 # assume nominal
    vel_incr = avg_accel * delta_t # (accel) velocity increase, ft/sec
    self.intacc_vel += vel_incr # new velocity, avg over last reading frame
    self.intacc_alt += self.intacc_vel * delta_t

    # Most sensors are read at 50 samples/sec; accel sensors at 200 Hz
    # e.g. Mars Snooper E18 16nov08
    # Time	G-axial	G-rad	Vcont	FER	Temp	Baro	Flight
    # 0.000	104	-240	2	-	(670)	(8528)	1
    # 0.005	104	-292	-	0	-	-	1
    # 0.010	96	-252	-	-	691	-	1
    # 0.015	104	-247	-	-	-	8454	1
    # 0.020	112	-246	1	-	-	-	1
    # The following three are 200 Hz
    self.accel_hist.append(self.filt_accel) # store filtered accel hist

    if self.FER_accel_apo_time == None: # G-sensor data is useless after apogee
      self.acc_alt.append(self.intacc_alt,sex_adj) # store integrated-accel-alt
      self.acc_vel.append(-self.intacc_vel,sex_adj) # and velocity
    # The following is sampled at 50 Hz, but I need to do the math every
    # time, because the liftoff-detect algorithm needs the numbers, and
    # that has to happen before I have determined the MUX phase.
    atm = float(fld[ATM_IX])/10000  # 10000 x pressure in atm (absolute)
    atm_adj = atm*NOM_VREF/self.Vref
    if atm_adj < 0: return # avoid math error
    baro_asl_alt = (1-atm_adj**0.190263)*288.15/0.00198122
    baro_agl_alt = baro_asl_alt - self.liftoff_alt

    # These can't get plotted, so don't worry about extra data points;
    # furthermore, they should go into .csv output right from the start
    if behavior['parrot_version']>=2: # Vcont got sampled this time
      self.vcont = int(fld[PYRO_CONT_IX])  # pyro continuity voltage
    else: # Parrot v.1 .. Vbat got sampled this time
      self.vbat = float(fld[VBAT_IX])/1000 # Battery voltage .. mV -> Volts

    if self.phase >= BOOST:
      # The following multiplexed sensors are sampled at 50 Hz
      if mod==3: # Baro sensor was the MUX choice this round
        # DECAY of .97 works well for flights I've analyzed, at the
        # initial sampling rate of 200Hz.  When the sample rate drops to 4/sec
        # after about five minutes into the flight, DECAY should decrease.
        # Curiously, not a regular 4 Hz, but one 200Hz mux cycle then a pause.
        temp_adj_asl_ft = self.temp_comp_baro(baro_asl_alt) + baro_asl_alt
        try: # Calculate delta-t: time tween prior two samples
          delta_t = sex_adj - self.alt.t[-1]
          if delta_t < .025: # under 50Hz period? (.02s)
            decay = DECAY
          else:
            decay = SLOW_DECAY
          self.filt_alt = self.filt_alt*decay + (1-decay)*baro_agl_alt
        except IndexError: # first time: self.alt has no data yet
          delta_t = 0
        # Filter axial-G data.  Subtract 1.0 G first, to account for gravity
        self.filt_accel = self.filt_accel*G_FILT_DECAY \
                           + (1-G_FILT_DECAY)*(g_ax_adj-1)
        if self.phase < COAST  and  self.filt_accel > self.max_accel:
          self.max_accel = self.filt_accel
        self.alt.append(baro_agl_alt,sex_adj) # Add to barometric alt history
        self.filt.append(self.filt_alt,sex_adj) # and to filtered alt history
        self.tcomp_agl = temp_adj_asl_ft - self.tcomp_liftoff_alt
        self.tcomp_alt.append(self.tcomp_agl,sex_adj)
        if self.filt_alt > self.baro_alt_max:
          self.baro_alt_max = self.filt_alt # latch new max altitude
          self.max_alt_time = sex_adj
      elif mod==1:
        if behavior['parrot_version']>=2: # FER got sampled this time
          FER = int(fld[FER_IX]) # Flight Event Register
          self.FER.append(FER,sex_adj)
      elif mod==2: # Temperature, this time
        self.temp.append(temp,sex_adj) # append to temperature history list
      # mod==0 (Vcont) handled independent of flight phase

      self.time_hist.append(sex_adj)  # 200 Hz time list
      max = self.baro_alt_max

      # Coast phase detection
      if self.phase < COAST  and  mod==3:
        if self.filt_accel < G_THRESH_COASTING:
          self.phase = COAST
          if log!=None: log("Phase -> Coasting, T+ %g sec" % sex_adj)
          self.burnout_time = sex_adj
          self.avg_thrust_accel = sum(self.accel_hist[self.liftoff_time_ix:])
          self.avg_thrust_accel /= float(len(self.accel_hist) \
                                          - self.liftoff_time_ix)
          # That's in Gs, not ft/sec/sec

      # Descent rate calculation, and arrived-at-ground test
      if mod==3:
        if len(self.filt.x) < self.desc_window:
          this_desc_window = len(self.filt.x) # not enough data yet
          delta_t = sex_adj - self.filt.t[-this_desc_window]
        else:
	  # If sampling rate has dropped from 50 Hz to 1 Hz,
          # the descent-rate window (number-of-samples) will need to shrink.
          delta_t = sex_adj - self.filt.t[-self.desc_window]
          while delta_t > DESC_WINDOW_SEC  and  self.desc_window > 1:
            self.desc_window -= 1 # will iterate 49x in two cases
            delta_t = sex_adj - self.filt.t[-self.desc_window]
          this_desc_window = self.desc_window
        if delta_t > 0:
          climb = self.filt_alt - self.filt.x[-this_desc_window] # alt. incr.
          # Apply another filter, then look for unchanging descent rate
          decay = DESC_RATE_DECAY
          if this_desc_window < 20: decay = DESC_RATE_SLOW_DECAY
          new_desc_rate = self.desc_rate*decay - (1-decay)*climb
          self.down = self.desc_rate/delta_t # Ft/sec
          # Only check whether Down when altitude is low (but not boosting)
          if sex_adj > DOWN_THRESH_SEC  and  self.filt_alt < DOWN_THESH \
              and math.fabs(self.down) < DOWN_DESC_RATE_THRESH:
            if self.phase != DOWN:
              if behavior['verbose'] > 0:
                if log!=None: log("Phase -> Down, T+ %g sec" % sex_adj)
              self.phase = DOWN
              try: # average the last ten baro-alt readings to get landing-alt
                self.landing_alt = sum(self.filt.x[-10:])/10.0
              except IndexError:
                self.landing_alt = self.filt_alt # shrug
            if behavior['proc_cutoff_downed']:
              return
            # otherwise, continue processing
          self.desc_rate = new_desc_rate
          self.desc.append(self.down,sex_adj) # was: new_desc_rate
      #end if mod==3:
    #end if self.phase >= BOOST:

    if FER != None:
      str = self.fer_events(FER,sex_adj)
      if str!=None and log!=None: log(str)

    fld.extend( ['']*(N_COLS-len(fld)))

    fld[SEX_ADJ_IX] = sex_adj
    fld[G_AX_ADJ_IX] = g_ax_adj
    fld[G_LAT_ADJ_IX] = g_lat_adj
    if behavior['parrot_version'] >= 2:
      fld[PYRO_CONT_IX] = self.vcont
    else:
      fld[VBAT_IX] = self.vbat
      fld[VREF_IX] = self.Vref
    fld[TEMP_IX] = temp
    fld[ATM_ADJ_IX] = atm_adj
    fld[BARO_ASL_ALT_IX] = baro_asl_alt
    fld[BARO_AGL_ALT_IX] = baro_agl_alt
    fld[MAX_IX] = max
    fld[FILT_ALT_IX] = self.filt_alt
    fld[DOWN_IX] = self.down
    fld[FILT_ACCEL_IX] = self.filt_accel
    fld[INTACC_VEL_IX] = self.intacc_vel
    try:
      fld[TCOMP_BARO_ALT] = self.tcomp_agl
    except:
      pass

    if self.phase >= BOOST:
      print1line(fld,outfile,behavior)
      return

    # Otherwise, we're in pre-launch phase
    self.pre_launch.append( fld)
  
    # Detect launch
    # Parrot v2 detects launch and reports it in Flight Event Register (FER)
    # It's the MSB of byte of 16-bit int, so the simple test is this: negative?
    if behavior['parrot_version'] >= 2 \
        and behavior['launch_detect_method'] == 'FER':
      if FER==None or FER==0:
        return # launch condition not satisfied
      if behavior['verbose']:
        if log!=None: log("Launch FER (%d) detected, T+ %g sec" % (FER,sex_adj))
      n = len(self.pre_launch)
      self.liftoff_time = sex_adj
    else: # Gotta auto-detect when no FER available, or when user says so
      # three consecutive axial G readings (unfiltered) over 3.5 => launch
      if len(self.pre_launch) <= 3:
        return # too early to test
      if self.pre_launch[-1][G_AX_ADJ_IX] < behavior['G_THRESH'] \
          or self.pre_launch[-2][G_AX_ADJ_IX] < behavior['G_THRESH'] \
          or self.pre_launch[-3][G_AX_ADJ_IX] < behavior['G_THRESH']:
        return # launch condition not satisfied
      if behavior['verbose'] and log!=None:
        log("Launch detected, %g sec (T+ 0)" % sex_adj)
      n = len(self.pre_launch) - 3 # exclude final 3 readings
      self.liftoff_time = self.pre_launch[-3][SEX_ADJ_IX]
    self.liftoff_time_ix = n
    # Launch detected .. Calculate average temp and barometric altitude ASL 
    for pre_fld in self.pre_launch[0:n]:
      self.liftoff_alt += pre_fld[BARO_ASL_ALT_IX]
    self.liftoff_alt /= n # average
    if self.ambient_pad_temp == 0: # it wasn't pre-set .. get temp from sensor
      for pre_fld in self.pre_launch[0:n]:
        self.ambient_pad_temp += pre_fld[TEMP_IX]
      self.ambient_pad_temp /= n # average
    self.delta_temp_from_std_atm = self.ambient_pad_temp \
                                    -59.0+STD_ATM_LAPSE_RATE*self.liftoff_alt
    # Calculate average temp-compensated barometric altitude ASL 
    for fld in self.pre_launch[0:n]:
      fld[TCOMP_BARO_ALT] = self.temp_comp_baro(fld[BARO_ASL_ALT_IX]) \
                             + fld[BARO_ASL_ALT_IX]
      self.tcomp_liftoff_alt += fld[TCOMP_BARO_ALT]
    self.tcomp_liftoff_alt /= n
    if behavior['verbose']>1 and log!=None:
      log("Temp-compensated avg pre-launch = %g ft ASL, avg temp = %g F" \
           % (self.tcomp_liftoff_alt,self.ambient_pad_temp))

    self.phase = BOOST
    header(outfile,behavior)

    # First off, figure out the A/D multiplexer phasing.
    # Most sensors are read at 50 samples/sec; accel sensors at 200 Hz
    # e.g. Mars Snooper E18 16nov08
    # Time	G-axial	G-rad	Vcont	FER	Temp	Baro	Flight
    # 0.000	104	-240	2	-	(670)	(8528)	1
    # 0.005	104	-292	-	0	-	-	1
    # 0.010	96	-252	-	-	691	-	1
    # 0.015	104	-247	-	-	-	8454	1
    # 0.020	112	-246	1	-	-	-	1
    prev_fld = self.pre_launch[0]
    n_samples = 1
    for fld in self.pre_launch[1:]:
      mux_offset = 0
      for ix in [VBAT_IX,VREF_IX,TEMP_IX,ATM_IX]:
        if fld[ix] != prev_fld[ix]:
          self.mux = ((n_samples - mux_offset - 1) % 4) + 1 # 1..4
          try: # Mysterious error, compiled .exe version only
           if behavior["verbose"]>1 and log!=None: log("mux = " + str(self.mux))
          except TypeError: pass #err("log = " + repr(log) )
          break
        mux_offset += 1
      prev_fld = fld
      n_samples += 1
    if not self.mux and err!=None:
      err("Couldn't sync with A/D multiplexer")

    # Now we can finally flush out the data we've been squirreling away
    # after performing fixup based on liftoff-altitude and -time
    for fld in self.pre_launch:
      fld[SEX_ADJ_IX] -= self.liftoff_time # i.e. sex_adj
      baro_agl_alt = fld[BARO_AGL_ALT_IX] - self.liftoff_alt
      fld[BARO_AGL_ALT_IX] = baro_agl_alt
      #self.alt.append(baro_agl_alt,sex_adj)
      # Await liftoff to start up digital filter on filt_alt; now still 0
      #self.filt_hist.append(self.filt_alt) # add to filtered alt history list
      self.time_hist.append(fld[SEX_ADJ_IX])  # ... and to parallel time list
      #self.accel_hist.append(self.filt_accel)
      fld[FILT_ALT_IX] = self.filt_alt
      try:
        self.tcomp_agl = fld[TCOMP_BARO_ALT] - self.tcomp_liftoff_alt
      except TypeError:  pass # fld[TCOMP_BARO_ALT] may be empty
      fld[TCOMP_BARO_ALT] = self.tcomp_agl
      print1line(fld,outfile,behavior)


    del self.pre_launch
    return


# ----- header ----------------------------------------------------------------
#

def header(outfile,behavior):

  """"Print header"""

  global N_COLS

  # Header-line listing each column's contents
  fld = ['']*N_COLS
  fld[SEX_IX] = "Raw"
  fld[SEX_ADJ_IX] = 'Tminus'
  fld[G_AX_ADJ_IX] = 'Axial'
  fld[G_LAT_ADJ_IX] = 'Lateral'
  if behavior['parrot_version'] >= 2:
    fld[PYRO_CONT_IX] = 'Cont'
    fld[VREF_IX] = 'FER'
  else:
    fld[VBAT_IX] = 'Bat'
    fld[VREF_IX] = 'Ref'
  fld[TEMP_IX] = 'Temp'
  fld[ATM_ADJ_IX] = 'Pressure'
  fld[BARO_ASL_ALT_IX] = 'Baro Alt'
  fld[BARO_AGL_ALT_IX] = 'Baro Alt'
  fld[MAX_IX] = 'Max Alt'
  fld[FILT_ALT_IX] = 'Filt Alt'
  fld[DOWN_IX] = 'Descent'
  fld[FILT_ACCEL_IX] = 'Accel'
  fld[FER_IX] = 'FER'
  fld[INTACC_VEL_IX] = 'Accel Vel'
  fld[TCOMP_BARO_ALT] = 'Temp-comp'
  print1line(fld,outfile,behavior)

  # Header-line listing units of measure
  fld = ['']*N_COLS
  fld[SEX_IX] = 't'
  fld[AXIAL_G_IX] = 'G/100'
  fld[LATERAL_G_IX] = 'G/100'
  if behavior['parrot_version'] >= 2:
    fld[PYRO_CONT_IX] = 'mV'
    fld[FER_IX] = ''
  else:
    fld[VBAT_IX] = 'V'
    fld[VREF_IX] = 'V'
  fld[ATM_IX] = 'Baro'
  fld[IND_IX] = 'flt'
  fld[SEX_ADJ_IX] = 'sec'
  fld[G_AX_ADJ_IX] = 'Gs'
  fld[G_LAT_ADJ_IX] = 'Gs'
  fld[TEMP_IX] = 'deg F'
  fld[ATM_ADJ_IX] = 'Atm'
  fld[BARO_ASL_ALT_IX] = 'ft ASL'
  fld[BARO_AGL_ALT_IX] = 'ft AGL'
  fld[MAX_IX] = 'ft AGL'
  fld[FILT_ALT_IX] = 'ft AGL'
  fld[DOWN_IX] = 'ft/sec'
  fld[FILT_ACCEL_IX] = 'G'
  fld[INTACC_VEL_IX] = 'ft/sec'
  fld[TCOMP_BARO_ALT] = 'ft AGL'
  print1line(fld,outfile,behavior)
